#include <string.h>
#include <stdarg.h>
#include <misc.h>
#include "tlmpweb.h"

static W_INT formdo ("formdo");

class FORM_INFO{
	/*~PROTOBEG~ FORM_INFO */
	/*~PROTOEND~ FORM_INFO */
};
enum FORM_STATUS {FORM_NOTHING, FORM_SETUP, FORM_VALIDATE, FORM_PROCESS};

struct _F_form_private{
	FORM_INFO info;
	FORM_STATUS status;
	SSTRING lastname;
	SSTRING lastvalue;
};



static const char *form_geteditval(
	const char *name,
	const char *value)
{
	if (t_var_isdef(name)){
		// We are re-editing after an error, so use the value
		// previously entered by a user
		value = t_varstr(name);
	}
	return value;
}


PUBLIC _F_form::_F_form()
{
	priv = new _F_form_private;
	priv->status = FORM_NOTHING;
}

PUBLIC void _F_form::fputs (const char *s, FILE *f)
{
	if (f == stdout){
		if (priv->status == FORM_SETUP || priv->status == FORM_VALIDATE){
			::fputs (s,f);
		}
	}else{
		::fputs (s,f);
	}
}

PUBLIC void _F_form::field_radio (
	const char *name,
	const char *value,
	bool selected,
	const char *opt)
{
	if (t_var_isdef(name)){
		// We are re-editing after an error, so use the value
		// previously entered by a user
		const char *val = t_varstr(name);
		selected = strcmp(val,value) == 0;
	}
	char tmpopt[strlen(opt)+9];
	sprintf (tmpopt,"%s %s",selected ? "checked" : "", opt);
	setfield (name,"radio",value,tmpopt);
}
PUBLIC void _F_form::field_checkbox (
	const char *name,
	bool selected,
	const char *opt)
{
	if (t_var_isdef(name)){
		// We are re-editing after an error, so use the value
		// previously entered by a user
		selected = t_varint(name) != 0;
	}
	char tmpopt[strlen(opt)+9];
	sprintf (tmpopt,"%s %s",selected ? "checked" : "", opt);
	setfield (name,"checkbox","1",tmpopt);
}
PUBLIC void _F_form::field_string (
	const char *name,
	const char *value,
	const char *opt)
{
	setfield (name,"text",value,opt);
}

PUBLIC void _F_form::field_hidden (
	const char *name,
	const char *value)
{
	setfield (name,"hidden",value,"");
}

PUBLIC void _F_form::field_textarea (
	const char *name,
	int rows,
	int cols,
	const char *opt)
{
	priv->lastname.setfrom (name);
	if (priv->status == FORM_SETUP){
		::printf ("<textarea name=%s rows=%d cols=%d>\n",name,rows,cols);
		if (t_var_isdef(name)){
			::printf ("%s",t_varstr(name));
		}
	}
}

PUBLIC void _F_form::field_textarea_body(const char *ctl, ...)
{
	if (priv->status == FORM_SETUP && !t_var_isdef(priv->lastname.get())){
		va_list list;
		va_start (list,ctl);
		vprintf (ctl,list);
		va_end (list);
	}
}
PUBLIC void _F_form::field_textarea_end()
{
	if (priv->status == FORM_SETUP){
		::printf ("</textarea>\n");
	}
}
PUBLIC void _F_form::field_list (
	const char *name,
	const char *value,
	const char *opt)
{
	if (priv->status == FORM_SETUP){
		::printf ("<select name=%s>\n",name);
		priv->lastvalue.setfrom (form_geteditval(name,value));
	}
}

PUBLIC void _F_form::field_list_item (
	const char *sel_value,
	const char *sel_desc)
{
	if (priv->status == FORM_SETUP){
		printf ("<option value=\"%s\" %s>%s\n",sel_value
			,priv->lastvalue.cmp(sel_value)==0 ? "selected" : ""
			,sel_desc);
	}
}

PUBLIC void _F_form::field_list_item (
	const char *sel_value)
{
	field_list_item(sel_value,sel_value);
}


PUBLIC void _F_form::field_list_end()
{
	if (priv->status == FORM_SETUP){
		::printf ("</select>\n");
	}
}


PUBLIC void _F_form::printf (const char *ctl, ...)
{
	if (priv->status == FORM_SETUP || priv->status == FORM_VALIDATE){
		va_list list;
		va_start (list,ctl);
		vprintf (ctl,list);
		va_end (list);
	}
}

PUBLIC void _F_form::setfield (
	const char *name,
	const char *type,
	const char *value,
	const char *opt)
{
	//tbvar->add (name
	if (priv->status == FORM_SETUP){
		value = form_geteditval(name,value);
		::printf ("<input type=%s name=%s value=\"%s\" %s>\n",type,name,value,opt);
	}
}

static void setup (_F_form &c, bool first)
{
	c.priv->status = FORM_SETUP;
	printf ("<form action=%s>\n",tlmpweb_curpage());
	printf ("<input type=hidden name=formdo value=1>\n");
	c.setup(first);
	printf ("<input type=submit>\n");
	printf ("</form>\n");
}

void form (_F_form &c)
{
	if (formdo == 0){
		// Prepare the form
		setup (c,true);
	}else if (formdo == 1){
		c.priv->status = FORM_NOTHING;
		c.setup(false);		// Call this to learn the definition of the fields
		c.priv->status = FORM_VALIDATE;
		// We can do some pre validation based on the type. We will do this
		// one day.
		// ...
		if (c.validate()){
			c.process();
		}else{
			setup (c,false);
		}
	}else{
		tlmpweb_error ("Invalid formdo value\n");
	}
}

