#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <dlfcn.h>
#include <misc.h>
#pragma implementation
#include "tlmpweb.h"
#include "tlmpweb.m"
#include <tlmplib.h>
#include <translat.h>

static SSTRINGS tbvar,tbval;

void tlmpweb_header(const char *type)
{
	static bool done = false;
	if (!done){
		done = true;
		printf ("Content-Type: %s\n",type);
		printf ("\n");
	}
}

static bool title_done = false;

void tlmpweb_title(const char *title, ...)
{
	title_done = true;
	tlmpweb_header ("text/html");
	printf ("<HTML>\n");
	printf ("<HEAD>\n");
	va_list list;
	va_start (list,title);
	printf ("<title>");
	vprintf (title,list);
	printf ("</title>");
	printf ("</head>\n");
}

void tlmpweb_body (const char *bgcolor, const char *vlink)
{
	tlmpweb_body (bgcolor,vlink,"");
}

void tlmpweb_body (const char *bgcolor, const char *vlink, const char *opt)
{
	printf ("<body");
	if (bgcolor != NULL) printf (" bgcolor=%s",bgcolor);
	if (vlink   != NULL) printf (" vlink=%s",vlink);
	printf (" %s>\n",opt);
}

void tlmpweb_error (const char *msg, ...)
{
	tlmpweb_header ("text/html");
	va_list list;
	va_start (list,msg);
	printf ("tlmpweb: ");
	vprintf (msg,list);
	va_end (list);
}

static const char *curdir;
/*
	Load a shared object either in the lib sub-directory of the
	current html directory or in the /usr/lib/tlmp/lib directory

	Exit if the lib can't be found.
*/
void tlmpweb_loadlib (const char *name)
{
	char path[PATH_MAX];
	snprintf (path,sizeof(path)-1,"%s/lib/%s.so",curdir,name);
	void *handle = dlopen (path,RTLD_LAZY|RTLD_GLOBAL);
	if (handle == NULL){
		snprintf (path,sizeof(path)-1,"/usr/lib/tlmp/lib/%s.so",name);
		void *handle = dlopen (path,RTLD_LAZY|RTLD_GLOBAL);
		if (handle == NULL){
			tlmpweb_error (MSG_U(E_NOLIB,"Can't find library %s.so (%s)\n")
				,name,dlerror());
			exit (-1);
		}else{
			void (*fct)() = (void (*)()) dlsym (handle,"tlmp_initmod");
			if (fct != NULL) (*fct)();
		}
	}
}

static void tlmpweb_loadlibs (const char *path)
{
	FILE *fin = fopen (path,"r");
	if (fin != NULL){
		char line[PATH_MAX];
		while (fgets(line,sizeof(line)-1,fin)!=NULL){
			strip_end (line);
			if (line[0] != '\0') tlmpweb_loadlib (line);
		}
		fclose (fin);
	}
}

/*
	Load all library enumerated in the .tlmplibs file of the current
	directory
*/
void tlmpweb_loadlibs ()
{
	char path[PATH_MAX];
	snprintf (path,sizeof(path)-1,"%s/.tlmplibs",curdir);
	if (file_exist (path)){
		tlmpweb_loadlibs (path);
	}else{
		const char *root = getenv ("DOCUMENT_ROOT");
		if (root != NULL){
			snprintf (path,sizeof(path)-1,"%s/.tlmplibs",root);
			if (file_exist(path)){
				tlmpweb_loadlibs (path);
			}
		}
	}
}
/*
	Copy the directory part of a path into "dir"
*/
static void tlmpweb_setdir (const char *path, char dir[PATH_MAX])
{
	strncpy (dir,path,PATH_MAX-1);
	dir[PATH_MAX-1] = '\0';
	char *pt = strrchr (dir,'/');
	if (pt != NULL) *pt = '\0';
}


static void tlmpweb_error0 (const char *msg)
{
	tlmpweb_header ("text/html");
	printf ("tlmpweb: %s",msg);
}

/*
	Return one web variable as a string
	is_def will be true if the variable was found.
*/
const char *t_varstr (const char *var, bool &is_def)
{
	const char *ret = "";
	int no = tbvar.lookup(var);
	is_def = false;
	if (no != -1){
		ret = tbval.getitem(no)->get();
		is_def = true;
	}
	return ret;
}
/*
	Return one web variable as a string
*/
const char *t_varstr (const char *var)
{
	bool is_def;
	return t_varstr (var,is_def);
}
/*
	Return true if a variable is defined
*/
bool t_var_isdef (const char *var)
{
	return tbvar.lookup(var) != -1;
}

/*
	Return one web variable as an int
*/
int t_varint (const char *var)
{
	return atoi(t_varstr(var));
}


/*
	Return the name of the current web page
*/
const char *tlmpweb_curpage()
{
	const char *ret = "unknown";
	static const char *p = NULL;
	if (p == NULL){
		p = getenv ("PATH_INFO");
		if (p != NULL){
			const char *pt = strrchr(p,'/');
			if (pt != NULL) p = pt+1;
		}
	}
	if (p != NULL) ret = p;
	return ret;
}


static char hextoi (char asc)
{
	return isdigit(asc) ? asc - '0' : (toupper(asc) - 'A') + 10;
}

/*
	Extract and decode one line
*/
static const char *http_decode (const char *str, char *buf, int sizbuf)
{
	char *pt = buf;
	char *endbuf = buf + sizbuf-1;
	while (*str != '\0' && *str != '\n' && pt < endbuf){
		if (*str == '%'){
			str++;
			*pt++ = hextoi(*str++) * 16 + hextoi(*str++);
		}else if (*str == '+'){
			*pt++ = ' ';
			str++;
		}else{
			*pt++ = *str++;
		}
	}
	*pt = '\0';
	strip_end (buf);
	if (*str == '\n') str++;
	return str;
}

static void http_parsevar ()
{
	const char *buf = getenv ("QUERY_STRING");
	if (buf == NULL) return;
	while (1){
		char *pt = strchr(buf,'=');
		if (pt == NULL){
			break;
		}else{
			*pt++ = '\0';
			char *end = strchr(pt,'&');
			if (end != NULL){
				*end++ = '\0';
			}
			char var[200];
			char val[2000];
			http_decode (buf,var,sizeof(var));
			http_decode (pt,val,sizeof(val));
			tbvar.add (new SSTRING(var));
			tbval.add (new SSTRING(val));
			if (end == NULL) break;
			buf = end;
		}
	}
}

static int tlmpweb_exec (const char *path)
{
	int ret = -1;
	void *handle = dlopen (path,RTLD_LAZY);
	if (handle == NULL){
		tlmpweb_error ("CGI error: dlopen (%s,RTLD_LAZY) failed (%s)\n"
			,path,dlerror());
	}else{
		{
			char dir[PATH_MAX];
			strncpy (dir,path,sizeof(dir)-1);
			char *pt = strrchr(dir,'/');
			if (pt != NULL){
				*pt = '\0';
				chdir (dir);
			}
		}
		void (*fct)() = (void (*)()) dlsym (handle,"webmain");
		if (fct == NULL){
			tlmpweb_error ("Invalid object: No webmain symbol\n");
		}else{
			http_parsevar();
			wvar_init();
			(*fct)();
			if (title_done){
				printf ("</body>\n</html>\n");
			}
			ret = 0;
		}
	}
	return ret;
}

void tlmpweb_copy (const char *file)
{
	FILE *fin = fopen (file,"r");
	if (fin == NULL){
		tlmpweb_error ("Can't copy file %s (%s)\n",file,strerror(errno));
	}else{
		char buf[1000];
		while (fgets (buf,sizeof(buf)-1,fin)!=NULL){
			fputs (buf,stdout);
		}
		fclose (fin);
	}
}

void tlmpweb_include (const char *file)
{
	const char *pt = strrchr(file,'.');
	if (pt == NULL){
		tlmpweb_copy (file);
	}else{
		if (strcmp(pt,".hc")==0){
			char cwd[PATH_MAX];
			getcwd(cwd,sizeof(cwd)-1);
			if (file[0] == '/'){
				tlmpweb_exec (file);
			}else{
				char path[PATH_MAX];
				snprintf (path,sizeof(path)-1,"%s/%s",cwd,file);
				tlmpweb_exec (path);
			}
			chdir (cwd);
		}else{
			tlmpweb_copy (file);
		}
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	//translat_load ("/usr/lib/tlmp/help","tlmp", "eng");
	linuxconf_loadlibmsg("/usr/lib/linuxconf-lib/help");
	tlmp_seterrorfct (tlmpweb_error0);
	#if 0
		FILE *fout = fopen ("/tmp/toto.log","w");
		if (fout != NULL){
			char cwd[1000];
			getcwd (cwd,sizeof(cwd)-1);
			fprintf (fout,"cwd=%s\n",cwd);
			for (int i=0; i<argc; i++){
				fprintf (fout,"argv[%d]=:%s:\n",i,argv[i]);
			}
			extern char **environ;
			for (int i=0; environ[i] != NULL; i++){
				fprintf (fout,"environ[%d]=:%s:\n",i,environ[i]);
			}
			fclose (fout);
		}
	#endif
	const char *request = getenv ("REQUEST_URI");
	if (request != NULL && strncmp(request,"/cgi-bin/tlmpweb/",17)==0){
		// Accessing an .hc page in cgi-bin
		char cwd[1000];
		getcwd (cwd,sizeof(cwd)-1);
		curdir = cwd;
		char page[PATH_MAX];
		snprintf (page,sizeof(page)-1,"%s/%s",cwd,request+17);
		char *pt = strchr (page,'?');
		if (pt != NULL) *pt = '\0';
		tlmpweb_loadlibs ();
		ret = tlmpweb_exec (page);
	}else{
		const char *path = getenv ("PATH_TRANSLATED");
		if (path == NULL){
			tlmpweb_error ("CGI error: No PATH_TRANSLATED variable\n");
		}else{
			char dir[PATH_MAX];
			tlmpweb_setdir (path,dir);
			curdir = dir;
			tlmpweb_loadlibs ();
			ret = tlmpweb_exec (path);
		}
	}
	return ret;
}

